# -*- coding: utf-8 -*-
## PloneSubscription
## Copyright (C)2005 Ingeniweb

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; see the file COPYING. If not, write to the
## Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
"""
This module implements generic functions for subscriptions
"""
__version__ = "$Revision: 1.80 $"
# $Source: /cvsroot/ingeniweb/PloneSubscription/SubscriptionTool.py,v $
# $Id: SubscriptionTool.py,v 1.80 2005/12/12 10:10:59 roeder Exp $
__docformat__ = 'restructuredtext'

# Python imports
import socket
import md5
from smtplib import SMTPException

# Zope imports
from Globals import InitializeClass
from AccessControl import ClassSecurityInfo, getSecurityManager
from AccessControl.SecurityManagement import newSecurityManager
from AccessControl.SecurityManager import setSecurityPolicy
from AccessControl.SecurityManagement import noSecurityManager
from DateTime import DateTime
from Products.MailHost.MailHost import MailHostError

# CMF imports
from Products.CMFCore.tests.base.security import OmnipotentUser
from Products.CMFCore.tests.base.security import PermissiveSecurityPolicy
from Products.CMFCore import CMFCorePermissions
from Products.CMFCore.utils import getToolByName, UniqueObject
from Products.CMFPlone.PropertiesTool import SimpleItemWithProperties

# Archetypes imports
from Products.Archetypes.public import *
from Products.Archetypes.Renderer import renderer

# Products imports
from Products.PloneSubscription.config import *
from Products.PloneSubscription import SubscriptionPermissions
from Products.PloneSubscription.interfaces.ISubscription import ISubscription
from Products.PloneSubscription.interfaces.ISubscriber import ISubscriber
# import from schema
from Products.PloneSubscription.content.schemata import FormAddGroupSubscriptionSchema


_MARKER = 'Marker'

TOOL_META_TYPE = 'SubscriptionTool'

# Subscription tool schema. Only reference field will be editable
schema = BaseFolderSchema.copy() + Schema((
    BooleanField(
        'show_actions',
        searchable=False,
        default=False,
        write_permission=SubscriptionPermissions.ManageSubscriptionContent,
        widget=BooleanWidget(
            visible={ 'view': 'visible', 'edit': 'visible',},
            label="Show actions",
            label_msgid="label_subscription_show_actions",
            description="Show actions permitting to add providers in the site.",
            description_msgid="help_subscription_show_actions",
            i18n_domain="plonesubscription",
            ),
        ),
    BooleanField(
        'hide_remove_button',
        searchable=False,
        default=True,
        write_permission=SubscriptionPermissions.ManageSubscriptionContent,
        widget=BooleanWidget(
            visible={ 'view': 'visible', 'edit': 'visible',},
            label="Hide remove button",
            label_msgid="label_subscription_hide_remove_button",
            description="Hides remove button in 'my subscriptions' if nothing is selected.",
            description_msgid="help_subscription_hide_remove_button",
            i18n_domain="plonesubscription",
            ),
        ),
    ReferenceField(
        'provider',
        multiValued=False,
        required=True,
        searchable=False,
        relationship='subscriptionProviderOf',
        allowed_types=('SubscriptionProvider',),
        read_permission=CMFCorePermissions.View,
        write_permission=SubscriptionPermissions.ManageSubscriptionContent,
        widget=ReferenceWidget(
            destination_types=['SubscriptionProvider',],
            addable=False,
            visible={ 'view': 'visible', 'edit': 'visible',},
            label="Subscription container provider",
            label_msgid="label_subscription_provider",
            description="Folder which provide base subscription functionnality.",
            description_msgid="help_subscription_provider",
            i18n_domain="plonesubscription",
            ),
        ),
    ReferenceField(
        'templates',
        multiValued=False,
        required=True,
        searchable=False,
        relationship='templatesProviderOf',
        allowed_types=('TemplatesProvider',),
        read_permission=CMFCorePermissions.View,
        write_permission=SubscriptionPermissions.ManageSubscriptionContent,
        widget=ReferenceWidget(
            destination_types=['TemplatesProvider',],
            addable=False,
            visible={ 'view': 'visible', 'edit': 'visible',},
            label="Templates provider",
            label_msgid="label_subscription_templates",
            description="Folder which provide base templates.",
            description_msgid="help_subscription_templates",
            i18n_domain="plonesubscription",
            ),
        ),
    IntegerField(
        'periodicity',
        required=True,
        searchable=True,
        default=60,
        vocabulary='getPeriodicityVocabulary',
        enforceVocabulary=True,
        read_permission=CMFCorePermissions.View,
        write_permission=SubscriptionPermissions.ManageSubscriptionContent,
        widget=SelectionWidget(
            label="Mail periodicity",
            label_msgid="label_subscription_periodicity",
            description="Periodicity used to send subscriptions mails.",
            description_msgid="help_subscription_periodicity",
            i18n_domain="plonesubscription",
            ),
        ),
    StringField(
        'base_url',
        required=False,
        searchable=True,
        default='',
        read_permission=CMFCorePermissions.View,
        write_permission=SubscriptionPermissions.ManageSubscriptionContent,
        widget=StringWidget(
            label="Base url",
            label_msgid="label_subscription_base_url",
            description="Site base url.",
            description_msgid="help_subscription_base_url",
            i18n_domain="plonesubscription",
            ),
        ),
    ReferenceField(
        'folderTemplate',
        multiValued=False,
        required=False,
        searchable=False,
        relationship='folderTemplateOf',
        allowed_types=('SubscriptionTemplate',),
        read_permission=CMFCorePermissions.View,
        write_permission=SubscriptionPermissions.ManageSubscriptionContent,
        widget=ReferenceWidget(
            destination_types=['SubscriptionTemplate',],
            addable=False,
            visible={ 'view': 'visible', 'edit': 'visible',},
            label="Folder subscription template",
            label_msgid="label_subscription_folder_template",
            description="Template for folder subscription.",
            description_msgid="help_subscription_folder_template",
            i18n_domain="plonesubscription",
            ),
        ),
    ReferenceField(
        'contentTemplate',
        multiValued=False,
        required=False,
        searchable=False,
        relationship='contentTemplateOf',
        allowed_types=('SubscriptionTemplate',),
        read_permission=CMFCorePermissions.View,
        write_permission=SubscriptionPermissions.ManageSubscriptionContent,
        widget=ReferenceWidget(
            destination_types=['SubscriptionTemplate',],
            addable=False,
            visible={ 'view': 'visible', 'edit': 'visible',},
            label="Content subscription template",
            label_msgid="label_subscription_content_template",
            description="Template for folder subscription.",
            description_msgid="help_subscription_content_template",
            i18n_domain="plonesubscription",
            ),
        ),
    ReferenceField(
        'keywordsTemplate',
        multiValued=False,
        required=False,
        searchable=False,
        relationship='keywordsTemplateOf',
        allowed_types=('SubscriptionTemplate',),
        read_permission=CMFCorePermissions.View,
        write_permission=SubscriptionPermissions.ManageSubscriptionContent,
        widget=ReferenceWidget(
            destination_types=['SubscriptionTemplate',],
            addable=False,
            visible={ 'view': 'visible', 'edit': 'visible',},
            label="Keywords subscription template",
            label_msgid="label_subscription_keywords_template",
            description="Template for keywords subscription.",
            description_msgid="help_subscription_keywords_template",
            i18n_domain="plonesubscription",
            ),
        ),
    ReferenceField(
        'exactSearchTemplate',
        multiValued=False,
        required=False,
        searchable=False,
        relationship='exactSearchTemplateOf',
        allowed_types=('SubscriptionTemplate',),
        read_permission=CMFCorePermissions.View,
        write_permission=SubscriptionPermissions.ManageSubscriptionContent,
        widget=ReferenceWidget(
            destination_types=['SubscriptionTemplate',],
            addable=False,
            visible={ 'view': 'visible', 'edit': 'visible',},
            label="Exact Search subscription template",
            label_msgid="label_subscription_exactsearch_template",
            description="Template for exact search subscription.",
            description_msgid="help_subscription_exactsearch_template",
            i18n_domain="plonesubscription",
            ),
        ),
    ),)
schema['id'].widget.visible = { 'view': 'visible', 'edit': 'hidden',}
schema['title'].widget.visible = { 'view': 'visible', 'edit': 'hidden',}
schema['description'].widget.visible = { 'view': 'visible', 'edit': 'hidden',}


class SubscriptionTool(UniqueObject, BaseFolder):
    """A tool to manage placeless subscription"""

    plone_tool = 1
    id = 'portal_subscription'
    title = "Placeless Subscription tool for Plone"
    portal_type = meta_type = TOOL_META_TYPE
    manage_options = BaseFolder.manage_options

    # AT specific
    archetype_name = "Subscription Tool"
    immediate_view = 'base_view'
    default_view   = 'base_view'
    global_allow = False
    allowed_content_types = ()
    schema = schema
    include_default_actions = 1

    # Tool specific
    security = ClassSecurityInfo()

    actions = (
        {
        'id'            : 'view',
        'name'          : 'View',
        'action'        : 'string:${object_url}/base_view',
        'condition'     : '',
        'permissions'   : (SubscriptionPermissions.ManageSubscriptionContent, ),
        'category'      : 'object',
        'visible'       : 1,
        },
        {
        'id'            : 'edit',
        'name'          : 'Edit',
        'action'        : 'string:${object_url}/base_edit',
        'condition'     : '',
        'permissions'   : (SubscriptionPermissions.ManageSubscriptionContent, ),
        'category'      : 'object',
        'visible'       : 1,
        },
    )

    # initialization

    def __init__(self):
        BaseFolder.__init__(self, self.id)

    # accessor and mutator overload

    security.declarePublic('showActions')
    def showActions(self):
        """give public accessor for show_actions"""
        return self.getShow_actions()

    security.declareProtected(SubscriptionPermissions.ManageSubscriptionContent, 'setPeriodicity')
    def setPeriodicity(self, periodicity, **kw):
        """Synchronise field data with subscription_task"""
        self.getField('periodicity').set(self, periodicity)

        mtool = getToolByName(self, 'portal_maintenance', None)
        if not mtool:
            return
        subscription_task = getattr(mtool.tasks, subscription_task_id)

        if periodicity == '60':
            subscription_task.scheduled_month = '*'
            subscription_task.scheduled_day_of_month = '*'
            #subscription_task.scheduled_day_of_week = '*'
            subscription_task.scheduled_hour = '*'
            subscription_task.scheduled_minute = '1'
        elif periodicity == '1440':
            subscription_task.scheduled_month = '*'
            subscription_task.scheduled_day_of_month = '*'
            #subscription_task.scheduled_day_of_week = '*'
            subscription_task.scheduled_hour = '1'
            subscription_task.scheduled_minute = '1'
        elif periodicity == '10080':
            subscription_task.scheduled_month = '*'
            subscription_task.scheduled_day_of_month = '*'
            #subscription_task.scheduled_day_of_week = '*'
            subscription_task.scheduled_hour = '1'
            subscription_task.scheduled_minute = '1'
        elif periodicity == '43200':
            subscription_task.scheduled_month = '*'
            subscription_task.scheduled_day_of_month = '1'
            #subscription_task.scheduled_day_of_week = '1'
            subscription_task.scheduled_hour = '1'
            subscription_task.scheduled_minute = '1'
        elif periodicity == '525600':
            subscription_task.scheduled_month = '1'
            subscription_task.scheduled_day_of_month = '1'
            #subscription_task.scheduled_day_of_week = '1'
            subscription_task.scheduled_hour = '1'
            subscription_task.scheduled_minute = '0'

    # methods
    security.declarePublic('hasSubscription')
    def hasSubscription(self, memberid=None):
        """Return True if the user has at least one subscription
           Used to decide whether the "my subscriptions" action is
           to be shown or not.
           Only works for anonymous currently, but could work with cookies.
        """
        has_subscriptions = False
        if not memberid:
            mtool = getToolByName(self, 'portal_membership')
            if not mtool.isAnonymousUser():
                member = mtool.getAuthenticatedMember()
                memberid = member.getId()
        subscriptions = self.getSubscriptions(subscriber_id=memberid)
        if len(subscriptions) > 0:
            has_subscriptions = True
        return has_subscriptions

    security.declarePublic('getSubscriberId')
    def getSubscriberId(self):
        subscriber_id = None
        mtool = getToolByName(self, 'portal_membership')
        if not mtool.isAnonymousUser():
            member = mtool.getAuthenticatedMember()
            subscriber_id = member.getId()
        return subscriber_id

    security.declarePublic('getSubscriber')
    def getSubscriber(self, subscriber_id=''):
        """Return subscriber where subscriptions are stored by default
           For Members, this will be a folder called like the member id
           For Anonymous, we can not return a subscriber.
        """
        # if no id is given, we assume that's the authenticated member's id
        subscriber=None
        if not subscriber_id:
            subscriber_id=self.getSubscriberId()
        if subscriber_id:
            provider = self.getProvider()
            subscriber = getattr(provider, subscriber_id, None)
        return subscriber

    security.declarePublic('isSubscribedTo')
    def isSubscribedTo(self, content, subscriber_id):
        """Checks whether the user has a subscription to the content"""
        is_subscribed_to = False
        subscriptions = self.getSubscriptions(content, subscriber_id)
        if len(subscriptions)>0:
            is_subscribed_to=True
        return is_subscribed_to

    security.declarePublic('getAnonymousSubscriberId')
    def getAnonymousSubscriberId(self, key):
        """
        Remove a subscription to the content.
        """
        return md5.md5(key).hexdigest()

    security.declarePublic('removeSubscription')
    def removeSubscription(self, content, subscriber_id=None):
        """
        Remove a subscription to the content.
        """
        is_anonymous=0
        mtool = getToolByName(self, 'portal_membership')
        if mtool.isAnonymousUser():
            is_anonymous=1
        if subscriber_id:
           pass 
        else:
            # Differentiate between anonymous and member subscribers for access
            if is_anonymous:
                email = self.REQUEST.get('email')
                subscriber_id = self.getAnonymousSubscriberId(email)
            else:
                member = mtool.getAuthenticatedMember()
                subscriber_id = member.getId()

        stool = getToolByName(self, 'portal_subscription')
        if is_anonymous:
            def getRolesInContext(self, context):
                return ['Manager',]
            def getRoles(self):
                return ['Manager',]
            OmnipotentUser.getRolesInContext = getRolesInContext
            OmnipotentUser.getRoles = getRoles
            portal = getToolByName(self, 'portal_url').getPortalObject()
            newSecurityManager(None, OmnipotentUser().__of__(portal.acl_users))

        subscriber = self.getSubscriber(subscriber_id)
        subscriptions = stool.getSubscriptions(content, subscriber_id)
        ids = []
        parent = None
        for subscription in subscriptions:
            subscription_object=subscription.getObject()
            if subscription_object:
                ids.append(subscription_object.getId())
                if not parent:
                    parent = subscription_object.getParentNode()
        if ids:
            parent.manage_delObjects(ids=ids)
            message = 'psm_content_removed_from_my_subscriptions'
        else:
            message = "No subscription found."

        if is_anonymous:
            message = 'psm_content_removed_from_my_subscriptions'
            noSecurityManager()

        return message

    security.declareProtected(SubscriptionPermissions.AddSubscriptionContent, 'setSubscription')
    def setSubscription(self, id='', ptype='', subscriber_uid='', **kw):
        """Create or modify a subscription for a user

        If you have a existing ptype implementing ISubscription we try to catch
        or create the object then we modify it with keywords argument.

        Keywords have to contain rpath and title key.
        """
        # Do not create anything if the portal type doesn't exist
        types_tool = getToolByName(self, 'portal_types')
        uid_catalog = getToolByName(self, 'uid_catalog')
        if ptype not in types_tool.objectIds():
            return None

        # Differentiate between anonymous and member subscribers for access
        mtool = getToolByName(self, 'portal_membership')
        if mtool.isAnonymousUser():
            subscriber_id = self.getAnonymousSubscriberId(kw['email'])
            subscriber_type = 'AnonymousSubscriber'
        else:
            member = mtool.getAuthenticatedMember()
            subscriber_id = member.getId()
            subscriber_type = 'UserSubscriber'

        # If UID is given, search in UID catalog or get the subscriber by id
        if subscriber_uid:
            subscriber = uid_catalog(UID=subscriber_uid)[0].getObject()
        else:
            subscriber = self.getSubscriber(subscriber_id)

        # Subscriber is not known, so add it, and get it
        if subscriber is None:
            provider = self.getProvider()
            provider.invokeFactory(subscriber_type, subscriber_id)
            subscriber = self.getSubscriber(subscriber_id)

        subscription = None
        if id == '':
            id = subscriber.generateUniqueId(ptype)
            subscription=None
        else:
            subscription = getattr(subscriber, id, None)

        if subscription:
            message = 'psm_subscriptions_updated'
        else:
            subscriber.invokeFactory(ptype, id)
            subscription = getattr(subscriber, id)
            if subscription:
                message = 'psm_subscriptions_page_added'
            else:
                message = 'psm_subscriptions_page_not_added'

        if not ISubscription.isImplementedBy(subscription):
            raise "Subscription interface not supported!"

        if kw.has_key('rpath'):
            subscription.setRpath(kw['rpath'])
        if kw.has_key('keywords'):
            subscription.setKeywords(kw['keywords'])
        if kw.has_key('title'):
            subscription.setTitle(kw['title'])
        if kw.has_key('folder'):
            subscription.setFolder(kw['folder'])
        if kw.has_key('content'):
            subscription.setContent(kw['content'])
        if kw.has_key('fullname'):
            subscription.setFullname(kw['fullname'])
        if kw.has_key('email'):
            subscription.setEmail(kw['email'])
        return message

    security.declarePublic('getAnonymousSubscription')
    def getAnonymousSubscription(self, content, subscriber_id):
        brefs = content.getBRefs(relationship='subscriptionContentOf')
        for subscription in brefs:
            if subscription.getSubscriber().getId()==subscriber_id:
                return subscription

    security.declarePublic('getSubscriptions')
    def getSubscriptions(self, content=None, subscriber_id=None):
        """Return Subscription stored in a SubscriptionFolder list in a dict
           Attention: This depends on the permissions of the user, as it is
                      using the catalog!
        """
        catalog = self.getCatalog()
        search_dict = {}
        if content:
            search_dict['rpath'] = '/'+content.absolute_url(relative=1)
        if subscriber_id:
            provider = self.getProvider()
            if provider:
                search_dict['path'] =   '/'+provider.absolute_url(relative=1)+'/'+subscriber_id
        if content:
            
            subscriptions = []
            for result in catalog.searchResults(search_dict):
                result_object = result.getObject()
                if result_object:
                    if result_object.getRefs(relationship='subscriptionContentOf'):
                        if result_object.getRefs(relationship='subscriptionContentOf')[0]==content:
                            subscriptions.append(result)
        else:
            subscriptions = []
            for result in catalog.searchResults(search_dict):
                result_object = result.getObject()
                if result_object:
                    subscriptions.append(result)
        return subscriptions

    security.declarePublic('countSubscriptions')
    def countSubscriptions(self, content):
        """Return the subscription count
        """
        count = 0
        brefs = content.getBRefs(relationship='subscriptionContentOf')
        if brefs:
            count=len(brefs)
        else:
            count=0
        return count

    security.declarePublic('getCatalog')
    def getCatalog(self):
        """Return subscription provider catalog"""
        catalog = None
        portal = getToolByName(self, 'portal_url').getPortalObject()
        catalog = getattr(portal,'subscription_catalog', None)
        return catalog

    security.declarePublic('getPeriodictyVocabulary')
    def getPeriodicityVocabulary(self):
        """Return strict periodicity vocabulary

        Use minutes for key to help DateTime operations
        """
        return Vocabulary(DisplayList((
            ('60', "Hourly", 'periodicity_hourly'),
            ('1440', "Daily", 'periodicity_daily'),
            # Not implemented in PloneMaintenance for the moment
            #('10080', "Weekly", 'periodicity_weekly'),
            ('43200', "Monthly", 'periodicity_monthly'),
            ('525600', "Yearly", 'periodicity_yearly'),
            )), self, 'plonesubscription')

    security.declarePublic('getGroupSubscriberVocabulary')
    def getGroupSubscriberVocabulary(self):
        """ Return a Vocabulary of created GroupSubscriber for adding group
            subscription interface
        """
        stool = getToolByName(self, 'portal_subscription')
        catalog = stool.getCatalog()
        uids_and_titles=[]
        for brain in catalog.searchResults(portal_type='GroupSubscriber'):
            obj = brain.getObject()
            if obj:
                uids_and_titles.append( (obj.UID(), obj.title_or_id() ) )
        return Vocabulary(DisplayList(uids_and_titles), self, 'plonesubscription')

    security.declareProtected(SubscriptionPermissions.ViewSubscriptionContent, 'getSubscriptionTemplate')
    def getSubscriptionTemplate(self, subscription=None):
        """get the template to use with a subscription"""
        ptype = subscription.portal_type

        if ptype == 'FolderSubscription' and self.getFolderTemplate() is not None:
            template = self.getFolderTemplate().getMail_body()

        if ptype == 'ContentSubscription' and self.getContentTemplate() is not None:
            template = self.getContentTemplate().getMail_body()

        elif ptype == 'KeywordsSubscription' and self.getKeywordsTemplate() is not None:
            template = self.getKeywordsTemplate().getMail_body()

        elif ptype == 'ExactSearchSubscription' and self.getExactSearchTemplate() is not None:
            template = self.getExactSearchTemplate().getMail_body()

        else:
            template = """
Your Subscription to "%(subscription_title)s" push you new data:

%(content)s
"""

        return template


    security.declareProtected(SubscriptionPermissions.ViewSubscriptionContent, 'getUpdatedSubscriptions')
    def getUpdatedSubscriptions(self, date):
        """Get all updated subscription since a date"""
        return tuple(self.getCatalog().searchResults(
            getLast_Send={
                'query': [date - float(self.getPeriodicity())/1440.0,],
                'range': 'max'
                },
            ))

    security.declareProtected(SubscriptionPermissions.ManageSubscriptionContent, 'mailing')
    def mailing(self):
        """get all information and send mails"""
        portal = getToolByName(self, 'portal_url').getPortalObject()
        portal_path = getToolByName(self, 'portal_url').getPortalPath()
        mtool = getToolByName(self, 'portal_membership')
        ptool = getToolByName(self, 'portal_properties')
        now = DateTime()

        # get mail headers data
        date = now.rfc822()
        email_from_address = portal.getProperty('email_from_address', 'admin@plone-subscription.com')
        charset = ptool.site_properties.getProperty('default_charset', 'utf-8')
        headers_infos = {
            'date': date,
            'mail_from': email_from_address,
            'mail_to': '',
            'subject': '',
            'charset': charset,
            }

        current_user = getSecurityManager().getUser()
        subscriptions = self.getUpdatedSubscriptions(now)
        portal_path_length = len(portal_path)
        for sub_brain in subscriptions:
            subscription = sub_brain.getObject()

            # Take care of subscription owner rights
            users = subscription.getSubscriber().getUsers(subscription.id)
            for user in users:
                newSecurityManager(None, user)

                brains = subscription.afterQueryFilters(subscription.getQueryBrains())

                # Restore current user permission
                newSecurityManager(None, current_user)

                template = self.getSubscriptionTemplate(subscription)
                if brains in ([], ()):
                    continue

                content = ''
                for brain in brains:
                    content += "    " + str(brain.Title or brain.getId) + " :\n"
                    base_url = self.getBase_url()
                    if base_url == '':
                        content += str(brain.getURL()) + "\n\n"
                    else:
                        content += base_url + str(brain.getURL(relative=1)[portal_path_length:]) + "\n\n"

                body =  template % {
                    'subscription_title': subscription.TitleOrId(),
                    'content': content,
                    }

                member = mtool.getMemberById(user.getId())
                # Check if member has subscribed and if he has an email
                if member is None:
                    # perhaps a user from Zope root acl_users
                    continue
                member_email = member.getProperty('email', '')
                if member_email == '':
                    continue

                headers_infos['mail_to'] = member_email
                headers_infos['subject'] = subscription.Title()

                self.sendMail(self.getHeaders(headers_infos) + "\n\n" + body)

            subscription.setLast_send(now)

        Log(LOG_DEBUG, "Mailing done")
        return 'mailing done'

    security.declareProtected(SubscriptionPermissions.ManageSubscriptionContent, 'mailing')
    def contactMailing(self, content, emails, from_email, subject, message):
        """Contact all subscribers concering some content"""
        ptool = getToolByName(self, 'portal_properties')
        charset = ptool.site_properties.getProperty('default_charset', 'utf-8')
        headers_infos = {
            'date': DateTime().rfc822(),
            'mail_from': from_email,
            'mail_to': '',
            'subject': subject,
            'charset': charset,
            }
        template = self.getContactTemplate()
        content_info = ''
        content_info += "    " + str(content.title_or_id()) + " :\n"
        portal = getToolByName(self, 'portal_url').getPortalObject()
        portal_path = getToolByName(self, 'portal_url').getPortalPath()
        portal_path_length = len(portal_path)
        base_url = self.getBase_url()
        if base_url == '':
            content_info += str(content.absolute_url()) + "\n\n"
        else:
            content_info += base_url + str(content.absolute_url(relative=1)[portal_path_length:]) + "\n\n"
        content_info += message + "\n\n"
        body =  template % {'subscription_title': content.title_or_id(),
                            'content': content_info,
                           }
        successes = 0
        for email in emails:
            headers_infos['mail_to'] = email
            if self.sendMail(self.getHeaders(headers_infos) + "\n\n" + body):
                successes += 1
        if successes == 0:
            message = "None of the %s messages were sent." % len(emails)
        elif successes < len(emails):
            message = "Only %s of %s mails sent successfully." % (successes, len(emails))
        else:
            message = "All %d messages sent successfully" % len(emails)
        Log(LOG_DEBUG, "Contact Mailing done")
        return message

    security.declarePublic('unsubscriptionMailing')
    def unsubscriptionMailing(self, content, email):
        """Contact all subscribers concering some content"""
        portal = getToolByName(self, 'portal_url').getPortalObject()
        ptool = getToolByName(self, 'portal_properties')
        charset = ptool.site_properties.getProperty('default_charset', 'utf-8')
        from_email = portal.getProperty('email_from_address', 'admin@plone-subscription.com')
        headers_infos = {
            'date': DateTime().rfc822(),
            'mail_from': from_email,
            'mail_to': '',
            'subject': 'Unsubscription from ' + str(content.title_or_id()),
            'charset': charset,
            }
        template = self.getUnsubscriptionTemplate()
        content_info = ''
        content_info += "    " + str(content.title_or_id()) + " :\n"
        portal = getToolByName(self, 'portal_url').getPortalObject()
        portal_path = getToolByName(self, 'portal_url').getPortalPath()
        portal_path_length = len(portal_path)
        base_url = self.getBase_url()
        if base_url == '':
            content_info += str(content.absolute_url()) + "\n\n"
        else:
            content_info += base_url + str(content.absolute_url(relative=1)[portal_path_length:]) + "\n\n"

        subscriber_id = self.getAnonymousSubscriberId(email)
        subscription = self.getAnonymousSubscription(content, subscriber_id)
        if subscription:
            removal_id=subscription.getId()
            Log(LOG_DEBUG, "SubscriptionTool.unsubscriptionMailing removal_id: %s\n\n" % removal_id)
            content_info += "To unscubscribe, click the folling link:" + "\n\n"
            content_info += content.absolute_url() + '/removeSubscription?id='+ subscriber_id + "\n\n"
            body = template % {'subscription_title': content.title_or_id(),
                                'content': content_info,
                               }
            successes = 0
            headers_infos['mail_to'] = subscription.getSubscriber().getSubscriberInfo()['email']
            if self.sendMail(self.getHeaders(headers_infos) + "\n\n" + body):
                successes += 1
                Log(LOG_DEBUG, "Message sent to " + headers_infos['mail_to'])
            if successes == 0:
                Log(LOG_DEBUG, "No message sent for content " + content.absolute_url() + ", subscriber_id :" + str(subscriber_id))
                #message = "An error occurred while trying to send unsubscription information."
                # Keep lying to dissimulate the fact that failed to send a mail as we know there is a subscription.
                message = "Email with unsubscription information sent."
            else:
                message = "Email with unsubscription information sent."
        else:
            # Keep lying to dissimulate the fact that we have no subscription for this email.
            message = "Email with unsubscription information sent."
            Log(LOG_DEBUG, "No message sent for content " + content.absolute_url() + ", subscriber_id :" + str(subscriber_id))
        return message

    security.declareProtected(SubscriptionPermissions.ViewSubscriptionContent, 'getSubscriptionTemplate')
    def getContactTemplate(self):
        """get the template to use with contacting all subscribers"""
        template = """A message concerning your subscription to "%(subscription_title)s":

%(content)s
"""
        return template

    security.declareProtected(SubscriptionPermissions.ViewSubscriptionContent, 'getUnsubscriptionTemplate')
    def getUnsubscriptionTemplate(self):
        """get the template to use with unsubscription"""
        template = """A message concerning your unsubscription from "%(subscription_title)s":

%(content)s
"""
        return template

    security.declareProtected(SubscriptionPermissions.ManageSubscriptionContent, 'getHeaders')
    def getHeaders(self, headers_infos={}):
        """ return mail headers

        take a dictionnary with five item:
        - date
        - mail_from
        - mail_to
        - subject
        - charset
        """

        headers = """Date: %(date)s
From: %(mail_from)s
To: %(mail_to)s
Subject: %(subject)s
Content-Type: text/plain; charset=%(charset)s
"""
        return headers % headers_infos

    security.declarePrivate('sendMail')
    def sendMail(self, messageText, mto=None, mfrom=None, subject=None, encode=None):
        """Send a preformatted mail"""
        mail_sent=True
        Log(LOG_DEBUG, "SubscriptionTool.sendMail : mail body\n\n %s" % messageText)
        portal = getToolByName(self, 'portal_url').getPortalObject()
        try:
            portal.MailHost.send(messageText, mto, mfrom, subject, encode)
        except (socket.error, MailHostError, SMTPException), error_msg:
            Log(LOG_NOTICE, "SubscriptionTool.sendMail",
                "Error while sending mail:\n %s" % error_msg)
            mail_sent=False
        return mail_sent

    # rendering utils

    security.declarePublic('renderEmptyWidget')
    def renderEmptyWidget(self, instance, field, mode, **kwargs):
        """Render empty widget from field"""

        accessor = field.getAccessor(instance)
        widget = field.widget
        field_name = field.getName()
        return renderer.render(field_name=field_name, mode=mode, widget=widget,
                               instance=instance, field=field, accessor=accessor,
                               **kwargs)

    security.declarePublic('getFormAddGroupedSubscriptionFields')
    def getFormAddGroupedSubscriptionFields(self):
        """ Return adding grouped subscription form schema fields """
        return FormAddGroupSubscriptionSchema.fields()

    security.declarePrivate('getQueryBrains')
    def getQueryBrains(self, last_send, query):
        """Return the list of brains concerned by the mailing depending on
           the last send time and the query"""
        brains = []
        if float(DateTime() - last_send) < float(self.getPeriodicity())/1440.0:
            stool = getToolByName(self, 'portal_subscription')
            catalog = stool.getCatalog()
            brains = catalog.searchResults(**query)
        return brains

    security.declarePrivate('afterQueryWorkflowFilter')
    def afterQueryWorkflowFilter(self, brains, last_send):
        """Filter the list of brains for a workflow change depending on the 
           last send time.
        """
        results = []
        wtool = getToolByName(self, 'portal_workflow')
        for brain in brains:
            history = wtool.getInfoFor(brain.getObject(), 'review_history', [])
            for transition in history:
                if transition.has_key('time') and transition['time'] > last_send:
                    results.append(brain)
                    break
        return results

registerType(SubscriptionTool, PROJECTNAME)

# These methods would be useful with Zope 2.8 to get types implementing some
# services interfaces.

def getInterfaceImplementedBy(interface):
    """Return a list of portal_types implementing an interface"""
    return tuple([type['name'] for type in listTypes()
                  if interface.isImplementedByInstancesOf(type['klass'])])

def getSubscriberAllowedContentTypes():
    """Return portal_types implementing ISubscription interface"""
    return getInterfaceImplementedBy(ISubscription)

def getProviderAllowedContentTypes():
    """Return portal_types implementing ISubscriber interface"""
    return getInterfaceImplementedBy(ISubscriber)
